"""
Camera modules.
TODO: write doc
"""
__title__ = "Cameras"

from .basecamera import BaseCamera
from .basevideo import BaseVideo
from .basespectrograph import BaseSpectrograph
from .dummycamera import DummyCamera
from .dummyspectrograph import DummySpectrograph
